from typing import Optional

from fittie.fitfile.profile.base_types import BaseType, BASE_TYPES


class FieldDescription:
    developer_data_index: int
    field_definition_number: int
    field_name: str
    base_type: BaseType
    native_message_number: Optional[int] = None
    native_field_number: Optional[int] = None
    units: Optional[str] = None

    def __init__(self, **kwargs):
        # Required fields
        self.developer_data_index = kwargs["developer_data_index"]
        self.field_definition_number = kwargs["field_definition_number"]
        self.field_name = kwargs["field_name"]
        self.base_type = BASE_TYPES[kwargs["fit_base_type_id"]]

        # Optional fields
        self.native_message_number = kwargs.get("native_mesg_num")
        self.native_field_number = kwargs.get("native_field_num")
        self.units = kwargs.get("units")

    def __str__(self) -> str:
        developer_index = f"{self.developer_data_index}_{self.field_definition_number}"
        return f"Field:{self.field_name=} {developer_index=}{self.units=}"

    def __repr__(self) -> str:
        return str(self)



try:
    from enum import StrEnum as _StrEnum
except ImportError:
    from enum import Enum as _Enum

    class _StrEnum(str, _Enum):
        def __str__(self) -> str:
            return self.value


class Endianness(_StrEnum):
    """Helper enum for representing Big (>, MSB) and Little (<, LSB) Endianness"""

    BIG = ">"
    LITTLE = "<"

    def __repr__(self) -> str:
        return self.name



from __future__ import annotations  # Added for type hints

import struct
from typing import TYPE_CHECKING, Optional

from fittie.utils.datastream import Streamable
from fittie.utils.exceptions import DecodeException
from fittie.fitfile.field_definitions import (
    FieldDefinition,
    DeveloperFieldDefinition,
    decode_field_definition,
    decode_developer_field_definition,
)
from fittie.utils.endianness import Endianness

if TYPE_CHECKING:
    from fittie.fitfile.records import RecordHeader


class DefinitionMessage:
    """
    Defines a local message type and associates it to a specific FIT message,
    and then designate the byte alignment and field contents of the
    upcoming data message.

    Definition messages have a fixed content of 5 bytes, and a variable content part
    which contains the field definitions (3 bytes per field)

    Fixed content:
    - Reserved: 1 Byte
    - Architecture: 1 Byte
    - Global message number: 2 bytes
    - Number of fields: 1 byte
    """

    header: "RecordHeader"
    endianness: str
    global_message_type: int
    number_of_fields: int
    field_definitions: list[FieldDefinition]
    number_of_developer_fields: int
    developer_field_definitions: list[DeveloperFieldDefinition]

    def __init__(
        self,
        header: "RecordHeader",
        endianness: str,
        global_message_type: int,
        field_definitions: list[FieldDefinition],
        developer_field_definitions: Optional[list[DeveloperFieldDefinition]] = None,
    ):
        self.header = header
        self.endianness = endianness
        self.global_message_type = global_message_type
        self.number_of_fields = len(field_definitions)
        self.field_definitions = field_definitions

        if developer_field_definitions is None:
            developer_field_definitions = []

        self.number_of_developer_fields = len(developer_field_definitions)
        self.developer_field_definitions = developer_field_definitions

    def get_developer_field_definition(
        self, data_index: int, number: int
    ) -> Optional[DeveloperFieldDefinition]:
        """
        Retrieve a DeveloperFieldDefinition by data index and number.
        If no definition is found, None will be returned
        """
        for definition in self.developer_field_definitions:
            if definition.number == number and definition.data_index == data_index:
                return definition

        return None

    def __str__(self) -> str:
        return (
            f"DefinitionMessage:{self.endianness=}{self.global_message_type=}"
            f"{self.field_definitions=}{self.developer_field_definitions=}"
        ).replace("self.", " ")

    def __repr__(self) -> str:
        return str(self)


def decode_definition_message(
    header: "RecordHeader", data: Streamable
) -> DefinitionMessage:
    if not header.is_definition_message:
        raise DecodeException(
            detail="tried to decode definition message with a non-definition message "
            "header",
            position=data.tell(),
        )

    try:
        (reserved,) = struct.unpack("B", data.read(1))

        if bool(reserved):
            raise DecodeException(
                detail="received invalid data for a definition message",
                position=data.tell(),
            )

        (architecture,) = struct.unpack("B", data.read(1))
        endianness = Endianness.BIG if bool(architecture) else Endianness.LITTLE
        (global_message_type,) = struct.unpack(f"{endianness}H", data.read(2))
        (number_of_fields,) = struct.unpack("B", data.read(1))

        field_definitions: list[FieldDefinition] = []
        developer_field_definitions: Optional[list[DeveloperFieldDefinition]] = None

        for field in range(number_of_fields):
            field_definitions.append(decode_field_definition(data))

        if header.is_developer_data:
            developer_field_definitions = []
            (number_of_developer_fields,) = struct.unpack("B", data.read(1))

            for field in range(number_of_developer_fields):
                developer_field_definitions.append(
                    decode_developer_field_definition(data)
                )

        return DefinitionMessage(
            header=header,
            endianness=endianness,
            global_message_type=global_message_type,
            field_definitions=field_definitions,
            developer_field_definitions=developer_field_definitions,
        )

    except struct.error as exc:
        raise DecodeException(
            detail="could not decode definition message with provided data",
            position=data.tell(),
        ) from exc



import os.path
from pathlib import Path
from typing import Protocol, Optional, BinaryIO, Any, Union

from fittie.fitfile.crc import apply_crc


class Streamable(Protocol):
    def read(self, size: Optional[int] = 1) -> bytes:
        ...

    def tell(self) -> int:
        ...

    def close(self) -> None:
        ...


class DataStream:
    """
    A thin wrapper around a BinaryIO

    It allows a path, file content or a BinaryIO/Streamable to be provided as
    initial value.

    TODO: include CRC check
    """

    _data: BinaryIO
    _path: Optional[Union[str, Path]]
    _calculated_crc: int

    should_calculate_crc: bool

    def __init__(self, value: Any):
        self.should_calculate_crc = True
        self._calculated_crc = 0

        if DataStream.is_file(value):
            self._data = value
        elif DataStream.is_path(value):
            self._path = value
        elif DataStream.is_streamable(value):
            self._data = value
        else:
            raise ValueError(
                f"unsupported value received as stream input: {type(value)}"
            )

    @property
    def calculated_crc(self) -> int:
        """Returns the calculated crc, or 0 if crc calculation is disabled"""
        return self._calculated_crc

    def reset_crc(self) -> None:
        """Resets the calculated crc back to 0"""
        self._calculated_crc = 0

    def read(self, size: Optional[int] = 1) -> bytes:
        """
        Reads the provided number of bits from the wrapped BinaryIO data

        If a crc should be calculated, the internal crc property will be calculated
        for each byte that was read.
        """
        if not self.should_calculate_crc:
            return self._data.read(size)

        value = self._data.read(size)

        for idx in range(0, size):
            self._calculated_crc = apply_crc(self._calculated_crc, value[idx])
        return value

    def tell(self) -> int:
        """Returns the current stream position"""
        return self._data.tell()

    def __enter__(self):
        if not hasattr(self, "_path"):
            return self

        self._data = open(self._path, "rb")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._data.close()

    @staticmethod
    def is_file(value) -> bool:
        """
        Checks if the provided value is a BufferedReader (file) with the correct mode.

        Raises IOError if file is not opened in rb mode
        """

        if not hasattr(value, "mode"):
            return False

        if (mode := getattr(value, "mode")) != "rb":
            raise IOError(f"expected file to be opened with mode 'rb', got '{mode}'")

        return True

    @staticmethod
    def is_path(value: Union[str, Path]) -> bool:
        """
        Check if the provided value is either a path string or Path, and checks if the
        file exists.

        Raises FileNotFoundError if path doesn't exist.
        """
        if not isinstance(value, str) and not isinstance(value, Path):
            return False

        if not os.path.exists(value):
            raise FileNotFoundError(f"file {value} does not exist")

        return True

    @staticmethod
    def is_streamable(value: Streamable) -> bool:
        """Check if the provided value has a read and tell method"""
        return hasattr(value, "read") and hasattr(value, "tell")



from io import BytesIO
from typing import BinaryIO

import pytest

from fittie.fitfile.header import Header, decode_header


@pytest.fixture
def header_bytes() -> bytes:
    return b"\x0e \xf1\x07pf\x00\x00.FIT\xef\x91"


@pytest.fixture
def header_data(header_bytes) -> BinaryIO:
    return BytesIO(header_bytes)


def test_decode_header(header_data) -> None:
    header = decode_header(header_data)

    assert header.length == 14
    assert header.profile_version == 2033
    assert header.protocol_version == 32
    assert header.data_size == 26224
    assert header.data_type == ".FIT"
    assert header.crc == 37359


def test_header_encode(header_bytes) -> None:
    header = Header(
        length=14,
        profile_version=2033,
        protocol_version=32,
        data_size=26224,
        data_type=".FIT",
        crc=37359,
    )

    assert header.encode() == header_bytes



import copy

from functools import lru_cache

from fittie.fitfile.profile.message_profile import MessageProfile
from fittie.fitfile.profile.messages import MESSAGES


@lru_cache(maxsize=None)
def get_message_profile(number: int) -> MessageProfile:
    """
    A cached helper method to retrieve data from messages.py as a MessageProfile by
    providing the message profile number.


    Notes:
        Replace lru_cache with cache when minimum Python version is 3.9
    """
    if number not in MESSAGES:
        raise ValueError(f'unknown message number "{number}"')

    raw = copy.deepcopy(MESSAGES[number])

    raw["fields"]: dict[int, MessageProfile.FieldProfile] = {
        k: MessageProfile.FieldProfile(**v) for k, v in raw["fields"].items()
    }

    return MessageProfile(**raw)



from __future__ import annotations  # Added for type hints

from typing import Any, Optional, TYPE_CHECKING

from fittie.utils.datastream import Streamable
from fittie.utils.exceptions import DecodeException
from fittie.fitfile.definition_message import DefinitionMessage
from fittie.fitfile.field_definitions import read_field, read_developer_field
from fittie.fitfile.field_description import FieldDescription
from fittie.fitfile.profile.util import get_message_profile


if TYPE_CHECKING:
    from fittie.fitfile.records import RecordHeader


class DataMessage:
    """
    Contains a local message type and populated data fields as described by the
    preceding definition message.

    There are two types of data message:
        Normal Data Message
        Compressed Timestamp Data Message

    Related DefinitionMessages and DataMessages share a local message type
    """

    header: "RecordHeader"
    fields: dict[str, Optional[Any]]

    def __init__(self, header: "RecordHeader", fields: dict[str, Optional[Any]]):
        self.header = header
        self.fields = fields

    def get_field(self, field_name: str) -> Optional[Any]:
        """Retrieve a field by key from fields"""
        return self.fields.get(field_name, None)

    @property
    def local_message_type(self) -> int:
        """Retrieves the local message type from the message header"""
        return self.header.local_message_type

    def __str__(self) -> str:
        return f"DataMessage:{self.fields}".replace("self.", " ")

    def __repr__(self) -> str:
        return str(self)


def decode_data_message(
    header: "RecordHeader",
    message_definition: DefinitionMessage,
    developer_data: dict[int, dict[str, dict[int, FieldDescription]]],
    data: Streamable,
) -> DataMessage:
    message_profile = get_message_profile(message_definition.global_message_type)

    fields: dict[str, Any] = {}

    for field in message_definition.field_definitions:
        field_profile = message_profile.fields[field.number]
        field_data = read_field(
            field_definition=field,
            # field_profile=field_profile,
            endianness=message_definition.endianness,
            data=data,
        )

        # TODO: components, accumulate etc (from field_profile?)
        fields[field_profile.field_name] = field_data

    if message_definition.developer_field_definitions and not developer_data:
        raise DecodeException(
            detail="definition message contains developer fields, "
            "but no field descriptions are provided",
            position=data.tell(),
        )

    for field in message_definition.developer_field_definitions:
        try:
            field_description = developer_data[field.data_index]["fields"][field.number]
        except KeyError:
            raise DecodeException(
                detail=f"no field description found for field {field}",
                position=data.tell(),
            )

        developer_field_definition = message_definition.get_developer_field_definition(
            data_index=field_description.developer_data_index,
            number=field_description.field_definition_number,
        )

        if not developer_field_definition:
            raise DecodeException(
                detail=f"no developer field definition found for field {field}",
                position=data.tell(),
            )

        field_data = read_developer_field(
            field_description=field_description,
            field_definition=developer_field_definition,
            endianness=message_definition.endianness,
            data=data,
        )

        fields[field_description.field_name] = field_data

    return DataMessage(header=header, fields=fields)



from __future__ import annotations  # Added for type hints

import struct
from typing import Optional, Any

from fittie.utils.datastream import Streamable
from fittie.utils.exceptions import DecodeException
from fittie.fitfile.data_message import decode_data_message
from fittie.fitfile.definition_message import (
    DefinitionMessage,
    decode_definition_message,
)


class RecordHeader:
    """
    A one byte header to determine of the record is either a DefinitionMessage,
    Normal Data Message or a Compressed Timestamp Data Message

    Normal header
    - Bit 0-3, Value 0-15: Local message type
    - Bit 4, Value 0: Reserved
    - Bit 5, Value 0 or 1: Message Type Specific (e.g. Developer Data Flag)
    - Bit 6, Value 0 or 1: Message Type, 1: Definition Message, 0: Data Message
    - Bit 7, Value 0, Header type: Normal header

    Compressed timestamp header
    - Bit 0-4, Value 0-31, Time offset
    - Bit 5-6, Value 0-3: Local Message Type
    - Bit 7, Value 1: Compressed timestamp header
    """

    is_definition_message: bool
    is_developer_data: bool
    local_message_type: int
    is_compressed_timestamp_message: bool
    time_offset: Optional[int]

    def __init__(
        self,
        is_definition_message: bool,
        is_developer_data: bool,
        local_message_type: int,
        is_compressed_timestamp_message: bool,
        time_offset: Optional[int] = None,
    ):
        self.is_definition_message = is_definition_message
        self.is_developer_data = is_developer_data
        self.local_message_type = local_message_type
        self.is_compressed_timestamp_message = is_compressed_timestamp_message
        self.time_offset = time_offset

    def __str__(self) -> str:
        return (
            f"RecordHeader:{self.is_definition_message=}{self.is_developer_data=}"
            f"{self.local_message_type=}{self.time_offset=}"
        ).replace("self.", " ")


def read_record_header(data: Streamable) -> RecordHeader:
    try:
        (value,) = struct.unpack("B", data.read(1))

        # Use a bit mask to get bit 7 to determine if this is a normal (0) or
        # compressed timestamp header (1)
        if is_compressed_timestamp_message := bool(value >> 7):
            is_definition_message = False
            is_developer_data = False

            # Shift 5 places to the right to get bits 7, 6 and 5
            # then apply mask 0b011 to get bits 6 and 5
            local_message_type = (value >> 5) & 0b011

            # Apply mask 0b11111 to get bit 4, 3, 2, 1 and 0
            time_offset = value & 0b111111
        else:
            # Apply mask 0b1000000 to get bit 6 to determine if the message
            # is a definition message
            is_definition_message = bool(value & 0b1000000)

            # Apply mask 0b100000 to determine if message contains developer data
            is_developer_data = bool(value & 0b100000)

            # Bit 4 is reserved and always 0, extra validity check
            if bool(value & 0b10000):
                raise DecodeException(
                    detail="invalid byte received for record header",
                    position=data.tell(),
                )

            # Apply mask 0b1111 to get bit 3, 2, 1, and 0
            local_message_type = value & 0b1111
            time_offset = None

        return RecordHeader(
            is_definition_message=is_definition_message,
            is_developer_data=is_developer_data,
            local_message_type=local_message_type,
            is_compressed_timestamp_message=is_compressed_timestamp_message,
            time_offset=time_offset,
        )

    except struct.error as exc:
        raise DecodeException(
            detail="could not decode record header with provided data",
            position=data.tell(),
        ) from exc


def read_record(
    record_header: RecordHeader,
    definition_message: Optional[DefinitionMessage],
    developer_data: dict[int, dict[str, Any]],
    data: Streamable,
) -> Any:  # TODO: add ABC Message type
    if record_header.is_compressed_timestamp_message:
        # TODO
        ...
        return

    if record_header.is_developer_data:
        # TODO: check if this works correctly
        return decode_definition_message(record_header, data)

    if record_header.is_definition_message:
        return decode_definition_message(record_header, data)

    # Record is a data message
    if definition_message is None:
        raise DecodeException(
            detail=f"did not receive local message definition for number "
            f"{record_header.local_message_type}",
            position=data.tell(),
        )

    message = decode_data_message(
        record_header, definition_message, developer_data, data
    )
    return message



import pytest

from fittie.fitfile.crc import calculate_crc, apply_crc


def test_calculate_crc():
    assert calculate_crc(b"\x0e D\x08-\x86\x00\x00.FIT") == 3484


@pytest.mark.parametrize(
    "crc,value,expected",
    [
        (0, 14, 50305),
        (50305, 32, 47109),
        (47109, 68, 12408),
        (12408, 8, 58417),
    ],
)
def test_apply_crc(crc, value, expected):
    assert apply_crc(crc, value) == expected



from fittie.utils.endianness import Endianness


def test_endianness():
    assert Endianness.BIG == ">"
    assert Endianness.LITTLE == "<"


def test_endianness__str():
    assert str(Endianness.BIG) == ">"
    assert str(Endianness.LITTLE) == "<"



from abc import ABC, abstractmethod


class IterableMixin(ABC):
    @property
    @abstractmethod
    def _iter_collection(self):
        """
        Defines which collection of data the __next__ method should use.

        Make sure to decorate this with @cached_property if it is an expensive
        calculation.

        Also make sure to add __dict__ to  __slots__ if slots are used and the
        property is cached, because  __dict__ has to be mutable.
        """
        ...

    def __next__(self):
        if self._iter_index >= len(self._iter_collection):
            # Remove _iter_collection value from cache, if it exists.
            # If _iter_collection is cached, then it should exist and be cleared
            # as soon as the collection has been fully iterated.
            if "_iter_collection" in self.__dict__:
                del self.__dict__["_iter_collection"]

            raise StopIteration

        value = self._iter_collection[self._iter_index]
        self._iter_index += 1
        return value

    def __iter__(self):
        self._iter_index = 0
        return self



from typing import Optional


class DecodeException(Exception):
    default_detail = "could not decode provided data"
    position: int

    def __init__(self, *, detail: Optional[str], position: int):
        if not detail:
            self.detail = self.default_detail
        else:
            self.detail = detail

        self.position = position

    def __str__(self):
        return f"{self.detail} at position {self.position}"



TABLE = (
    0x0000,
    0xCC01,
    0xD801,
    0x1400,
    0xF001,
    0x3C00,
    0x2800,
    0xE401,
    0xA001,
    0x6C00,
    0x7800,
    0xB401,
    0x5000,
    0x9C01,
    0x8801,
    0x4400,
)


def apply_crc(crc: int, value: int) -> int:
    """
    Applies the crc computation on the provided value and returns it
    """
    # Compute checksum of lower four bits of byte
    tmp = TABLE[crc & 0xF]
    crc = (crc >> 4) & 0x0FFF
    crc = crc ^ tmp ^ TABLE[value & 0xF]

    # Compute checksum of upper four bits of byte
    tmp = TABLE[crc & 0xF]
    crc = (crc >> 4) & 0x0FFF
    crc = crc ^ tmp ^ TABLE[(value >> 4) & 0xF]

    return crc


def calculate_crc(data: bytes) -> int:
    """
    Calculates crc checksum for the entire provided data

    Compute method from https://developer.garmin.com/fit/protocol/
    """

    crc = 0

    for byte in data:
        crc = apply_crc(crc, byte)

    return crc



import io
from unittest.mock import patch

import pytest
import tempfile

from fittie.utils.datastream import DataStream
from fittie.fitfile.crc import apply_crc


def test_datastream_file():
    with tempfile.NamedTemporaryFile("rb") as file:
        DataStream(file)  # Should not raise
        assert DataStream.is_file(file)


def test_datastream_file__invalid_mode():
    with tempfile.NamedTemporaryFile("r") as file:
        with pytest.raises(IOError) as exc_info:
            DataStream(file)

    assert "expected file to be opened with mode 'rb', got 'r'" in str(exc_info.value)


def test_datastream_path():
    with tempfile.NamedTemporaryFile("rb") as file:
        DataStream(file.name)  # Should not raise
        assert DataStream.is_path(file.name)


def test_datastream_path__file_does_not_exist():
    with pytest.raises(FileNotFoundError) as exc_info:
        DataStream("/does/not/exist.fit")

    assert "file /does/not/exist.fit does not exist" in str(exc_info.value)


def test_datastream_streamable():
    data = io.BytesIO(b"123")
    DataStream(data)  # Should not raise
    assert DataStream.is_streamable(data)


def test_datastream_invalid_value():
    with pytest.raises(ValueError) as exc_info:
        DataStream(bytes([1, 2, 3]))

    assert "unsupported value received as stream input: <class 'bytes'>" in str(
        exc_info.value
    )


def test_datastream_crc():
    datastream = DataStream(io.BytesIO(b"123"))
    assert datastream.calculated_crc == 0

    with patch(
        "fittie.utils.datastream.apply_crc", side_effect=apply_crc
    ) as patched_apply_crc:
        datastream.read()

    assert datastream.calculated_crc != 0
    patched_apply_crc.assert_called_once()


def test_datastream_crc__crc_disabled():
    datastream = DataStream(io.BytesIO(b"123"))
    datastream.should_calculate_crc = False
    assert datastream.calculated_crc == 0

    with patch(
        "fittie.utils.datastream.apply_crc", side_effect=apply_crc
    ) as patched_apply_crc:
        datastream.read()

    assert datastream.calculated_crc == 0
    patched_apply_crc.assert_not_called()



from __future__ import annotations  # Added for type hints

import struct
from typing import Any

from fittie.utils.datastream import Streamable
from fittie.utils.exceptions import DecodeException
from fittie.fitfile.field_description import FieldDescription
from fittie.fitfile.profile.base_types import BaseType, BASE_TYPES


class DeveloperFieldDefinition:
    """
    Specifies Developer Data Fields which are used to map data within a DataMessage
    to the appropriate meta-data

    Byte 0 Field Number
    Byte 1 Size
    Byte 2 Developer Data Index
    """

    number: int
    size: int
    data_index: int

    def __init__(self, number: int, size: int, data_index: int):
        self.number = number
        self.size = size
        self.data_index = data_index

    def __str__(self) -> str:
        return (
            f"DeveloperFieldDefinition:{self.number=}{self.size=}{self.data_index=}"
        ).replace("self.", " ")

    def __repr__(self) -> str:
        return str(self)


class FieldDefinition:
    """
    Specifies which FIT fields are included in the upcoming data message

    Byte 0 Field Definition Number
    Byte 1 Size
    Byte 2 Base Type
    """

    number: int
    size: int
    base_type: BaseType

    def __init__(self, number: int, size: int, base_type: BaseType):
        self.number = number
        self.size = size
        self.base_type = base_type

    def __str__(self) -> str:
        return f"FieldDefinition:{self.number=}{self.size=}{self.base_type=}".replace(
            "self.", " "
        )

    def __repr__(self) -> str:
        return str(self)


def decode_developer_field_definition(data: Streamable) -> DeveloperFieldDefinition:
    """
    Decode data into a DeveloperFieldDefinition
    """
    try:
        number, size, data_index = struct.unpack("3B", data.read(3))
    except struct.error as exc:
        raise DecodeException(
            detail="could not decode developer field definition with provided data",
            position=data.tell(),
        ) from exc

    return DeveloperFieldDefinition(
        number=number,
        size=size,
        data_index=data_index,
    )


def decode_field_definition(data: Streamable) -> FieldDefinition:
    """
    Decode data into a FieldDefinition

    If number equals 255 a DecodeException will be raised
    If no base type can be found for the base type number, a DecodeException will be
    raised
    """
    try:
        number, size, base_type_number = struct.unpack("3B", data.read(3))
    except struct.error as exc:
        raise DecodeException(
            detail="could not decode field definition with provided data",
            position=data.tell(),
        ) from exc

    if number == 255:
        raise DecodeException(
            detail="invalid field definition number received",
            position=data.tell(),
        )

    if not (base_type := BASE_TYPES.get(base_type_number)):
        raise DecodeException(
            detail=f"invalid base type number received: {base_type_number=}",
            position=data.tell(),
        )

    return FieldDefinition(number=number, size=size, base_type=base_type)


def _retrieve_value(
    number_of_values: int,
    base_type: BaseType,
    endianness: str,
    data: Streamable,
) -> Any:
    if number_of_values > 1:
        value = []

        for n in range(number_of_values):
            # (n_value,) = struct.unpack(
            #     f"{endianness}{base_type.size}{base_type.fmt}",
            #     data.read(base_type.size)
            # )
            n_value = base_type.get_value(endianness, data)

            # Null terminated string check
            if base_type.value_type == str and n_value == b"\x00":
                continue  # Just continue, there can be more than one \x00

            if n_value == base_type.invalid_value:
                value.append(None)
            else:
                value.append(n_value)

        if base_type.value_type == str:
            value = b"".join(value).decode("utf-8")
    elif base_type.name == "string":
        # TODO: check if we can reach this separate from a multipart value
        ...
        value = None
    else:
        value = base_type.get_value(endianness, data)

    return value


def read_field(
    field_definition: FieldDefinition,
    endianness: str,
    data: Streamable,
) -> Any:
    """
    Read field by field definition
    """
    base_type = field_definition.base_type
    number_of_values = int(field_definition.size / base_type.size)

    return _retrieve_value(number_of_values, base_type, endianness, data)


def read_developer_field(
    field_description: FieldDescription,
    field_definition: DeveloperFieldDefinition,
    endianness: str,
    data: Streamable,
) -> Any:
    # TODO: move this to field_definitions
    """
    Read developer data field by field definition and field description
    """

    base_type = field_description.base_type
    number_of_values = int(field_definition.size / base_type.size)

    return _retrieve_value(number_of_values, base_type, endianness, data)



import pytest

from io import BytesIO

from fittie.utils.exceptions import DecodeException
from fittie.fitfile.definition_message import (
    decode_definition_message,
    DefinitionMessage,
)
from fittie.fitfile.field_definitions import DeveloperFieldDefinition
from fittie.fitfile.records import RecordHeader
from fittie.utils.endianness import Endianness


def test_decode_definition_message_example_record_1():
    """Example from Garmin FIT Protocol documentation"""
    header = RecordHeader(
        is_definition_message=True,
        is_compressed_timestamp_message=False,
        is_developer_data=False,
        local_message_type=0,
    )

    reserved = (0).to_bytes(length=1, byteorder="little")
    architecture = (0).to_bytes(length=1, byteorder="little")
    global_message_number = (0).to_bytes(length=2, byteorder="little")
    number_of_fields = (5).to_bytes(length=1, byteorder="little")

    # Field 0
    field_0 = bytearray([0, 1, 0])
    field_1 = bytearray([1, 2, 132])
    field_2 = bytearray([2, 2, 132])
    field_3 = bytearray([3, 4, 140])
    field_4 = bytearray([4, 4, 134])

    data = BytesIO(
        reserved
        + architecture
        + global_message_number
        + number_of_fields
        + field_0
        + field_1
        + field_2
        + field_3
        + field_4
    )
    definition_message = decode_definition_message(header, data)

    assert definition_message.global_message_type == 0
    assert len(definition_message.field_definitions) == 5
    assert definition_message.endianness == Endianness.LITTLE
    assert definition_message.developer_field_definitions == []
    assert definition_message.number_of_fields == 5


def test_decode_definition_message_invalid_record_header():
    header = RecordHeader(
        is_definition_message=False,  # set to False
        is_compressed_timestamp_message=False,
        is_developer_data=False,
        local_message_type=0,
    )

    data = BytesIO()

    with pytest.raises(DecodeException) as excinfo:
        decode_definition_message(header, data)

    assert (
        "tried to decode definition message with a non-definition message header"
        in str(excinfo.value)
    )


def test_decode_definition_message_invalid_reserved_bit():
    header = RecordHeader(
        is_definition_message=True,
        is_compressed_timestamp_message=False,
        is_developer_data=False,
        local_message_type=0,
    )

    data = BytesIO(bytearray([1]))

    with pytest.raises(DecodeException) as excinfo:
        decode_definition_message(header, data)

    assert "received invalid data for a definition message" in str(excinfo.value)


def test_get_developer_field_definition():
    developer_field_definition = DeveloperFieldDefinition(
        data_index=1, number=2, size=1
    )
    definition_message = DefinitionMessage(
        header=RecordHeader(
            is_definition_message=True,
            is_developer_data=False,
            is_compressed_timestamp_message=False,
            local_message_type=0,
        ),
        endianness=Endianness.LITTLE,
        global_message_type=0,
        field_definitions=[],
        developer_field_definitions=[developer_field_definition],
    )

    assert (
        definition_message.get_developer_field_definition(data_index=1, number=2)
        == developer_field_definition
    )

    assert (
        definition_message.get_developer_field_definition(data_index=1, number=1)
        is None
    )



MESG_NUMS = {
    0: "file_id",
    1: "capabilities",
    2: "device_settings",
    3: "user_profile",
    4: "hrm_profile",
    5: "sdm_profile",
    6: "bike_profile",
    7: "zones_target",
    8: "hr_zone",
    9: "power_zone",
    10: "met_zone",
    12: "sport",
    15: "goal",
    18: "session",
    19: "lap",
    20: "record",
    21: "event",
    23: "device_info",
    26: "workout",
    27: "workout_step",
    28: "schedule",
    30: "weight_scale",
    31: "course",
    32: "course_point",
    33: "totals",
    34: "activity",
    35: "software",
    37: "file_capabilities",
    38: "mesg_capabilities",
    39: "field_capabilities",
    49: "file_creator",
    51: "blood_pressure",
    53: "speed_zone",
    55: "monitoring",
    72: "training_file",
    78: "hrv",
    80: "ant_rx",
    81: "ant_tx",
    82: "ant_channel_id",
    101: "length",
    103: "monitoring_info",
    105: "pad",
    106: "slave_device",
    127: "connectivity",
    128: "weather_conditions",
    129: "weather_alert",
    131: "cadence_zone",
    132: "hr",
    142: "segment_lap",
    145: "memo_glob",
    148: "segment_id",
    149: "segment_leaderboard_entry",
    150: "segment_point",
    151: "segment_file",
    158: "workout_session",
    159: "watchface_settings",
    160: "gps_metadata",
    161: "camera_event",
    162: "timestamp_correlation",
    164: "gyroscope_data",
    165: "accelerometer_data",
    167: "three_d_sensor_calibration",
    169: "video_frame",
    174: "obdii_data",
    177: "nmea_sentence",
    178: "aviation_attitude",
    184: "video",
    185: "video_title",
    186: "video_description",
    187: "video_clip",
    188: "ohr_settings",
    200: "exd_screen_configuration",
    201: "exd_data_field_configuration",
    202: "exd_data_concept_configuration",
    206: "field_description",
    207: "developer_data_id",
    208: "magnetometer_data",
    209: "barometer_data",
    210: "one_d_sensor_calibration",
    225: "set",
    227: "stress_level",
    258: "dive_settings",
    259: "dive_gas",
    262: "dive_alarm",
    264: "exercise_title",
    268: "dive_summary",
    285: "jump",
    317: "climb_pro",
    375: "device_aux_battery_info",
    65280: "mfg_range_min",
    65534: "mfg_range_max",
}



from __future__ import annotations  # added for type hints

from dataclasses import dataclass
from typing import Union, Optional


@dataclass
class MessageProfile:
    @dataclass
    class FieldProfile:
        field_name: str
        field_type: str
        array: Optional[Union[bool, str, int]] = None
        components: Optional[Union[str, list[str]]] = None
        scale: Optional[Union[int, list[int]]] = None
        offset: Optional[int] = None
        units: Optional[str] = None
        bits: Optional[Union[int, str]] = None
        accumulate: Optional[Union[int, list[int]]] = None
        ref_field_name: Optional[Union[str, list[str]]] = None
        ref_field_value: Optional[Union[str, list[str]]] = None
        comment: Optional[str] = None

        @property
        def is_array(self) -> bool:
            """Check whether the field is an array"""
            return self.array is not None

        @property
        def has_components(self) -> bool:
            """Check whether the field has components"""
            return self.components is not None

    name: str
    fields: dict[int, FieldProfile]
    group: str



from datetime import datetime, timezone

from fittie.fitfile.util import rollover_timestamp, datetime_from_timestamp


def test_rollover_timestamp():
    """
    Test data from FIT file documentation
    """
    initial_timestamp = 0x0000003B

    assert (
        resultant_timestamp := rollover_timestamp(initial_timestamp, 0b11101)
    ) == 0x3D

    assert (
        resultant_timestamp := rollover_timestamp(resultant_timestamp, 0b00010)
    ) == 0x42

    assert (
        resultant_timestamp := rollover_timestamp(resultant_timestamp, 0b00101)
    ) == 0x45

    assert rollover_timestamp(resultant_timestamp, 0b00001) == 0x61


def test_datetime_from_timestamp():
    assert datetime_from_timestamp(1046114793) == datetime(
        2023, 2, 23, 19, 26, 33, tzinfo=timezone.utc
    )
    assert datetime_from_timestamp(1046119077) == datetime(
        2023, 2, 23, 20, 37, 57, tzinfo=timezone.utc
    )
